#! /usr/bin/python
# Copyright (C) 2006, 2007 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Main entry point for the InteProxy"""

import os
import sys
import optparse
import urlparse
import urllib2
import locale
import inteproxy.proxyconnection as proxyconnection
import inteproxy.transcoder
import inteproxy.resources
from inteproxy.proxycore import (InteProxyHTTPRequestHandler,
                                 MasterWorkerServer,
                                 log_date_time_string)

inteproxy_version = "0.3.0"


class HTTPRedirectHandler(urllib2.HTTPRedirectHandler):

    """A version of urllib2's HTTPRedirectHandler that does not redirect

    The class urllib2.HTTPRedirectHandler implements error handlers for
    the redirect error codes sent by the server and then starts a new
    request.  In InteProxy we want to let InteProxy's client to handle
    the redirects themselves, so we overrride the error handling and do
    not handle any of the errors.  urrlib2 then falls back on the real
    http error handler so that InteProxy handles them like other errors
    such as 404.
    """

    def http_error_302(self, *args):
        """Overrides the inherited method and simply returns None.
        None indicates to urllib2 that the method could not handle the
        error and that urllib2 should look for another handler.
        """
        return None
    http_error_301 = http_error_303 = http_error_307 = http_error_302


def setup_urllib2(debuglevel):
    handlers = []
    if os.environ.has_key('https_proxy'):
        https_proxy = urlparse.urlsplit(os.getenv('https_proxy'))[1]
        del os.environ['https_proxy']
        sys.stderr.write("[%s] Using HTTPS proxy: %s\n"
                         % (log_date_time_string(), https_proxy))
        handlers.extend((
            proxyconnection.ConnectHTTPHandler(proxy=https_proxy,
                                               debuglevel=debuglevel),
            proxyconnection.ConnectHTTPSHandler(proxy=https_proxy,
                                                debuglevel=debuglevel)))
    else:
        httphandler = urllib2.HTTPHandler(debuglevel=debuglevel)
        httpshandler = urllib2.HTTPSHandler(debuglevel=debuglevel)
        handlers.extend((httphandler, httpshandler))

    handlers.append(HTTPRedirectHandler)
    urllib2.install_opener(urllib2.build_opener(*handlers))


def setup_logging(opts):
    """Sets up logging according to opts.

    Log output in InteProxy is written to sys.stderr.  This function
    redirects stderr to the appropriate place.

    The opts parameter should be the options object returned by
    optparse.  In particular, opts.logfile is used to determine where to
    log to.  It should either be None, so that the default redirection
    is done, or a filename in which case sys.stderr is changed to write
    to that file.

    The default behavior is trying to accomodate the following
    requirements:

     - When run directly from the source tree in a terminal, the log out
       should simply go to stderr.  That is, sys.stderr is not changed
       at all.

     - When run from an executable created with py2exe, no logging takes
       place, to avoid py2exe popping up a message box indicating that
       errors occurred.  In this case sys.stderr is redirected to
       os.devnull.
    """
    if opts.logfile:
        sys.stderr = open(opts.logfile, "a", 0)
    elif inteproxy.resources.in_py2exe():
        sys.stderr = open(os.devnull, "w")

def setup_language():
    """Sets the language environment if it's not set yet.
    This function checks if any of the common language environment
    variables is already set and if none are set, it sets LANG from the
    default locale as determined by locale.getdefaultlocale.

    This is necessary on Windows because there the locale is not usually
    set through environment variables, but Python's gettext module uses
    the environment variables to determine which language to use.
    """
    for envvar in ('LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG'):
        if envvar in os.environ:
            break
    else:
        default_locale = locale.getdefaultlocale()
        # The default is normally a tuple of two strings.  It may
        # contain None, objects under some circumstances, though.
        if len(default_locale) > 1:
            lang = default_locale[0]
            if isinstance(lang, str):
                os.environ["LANG"] = lang


# The main application object.  It's a global variable so that worker
# threads can easily get access to it.
the_application = None


def run_server(HandlerClass = InteProxyHTTPRequestHandler,
               ServerClass = MasterWorkerServer):
    """Run the InteProxy server"""
    parser = optparse.OptionParser()
    # The default config file name.  It's an absolute filename for the
    # default config file, so that it must be in a specific directory
    # (for now the same directory as InteProxy.py) and we don't
    # accidentally pick up other config files if InteProxy is started
    # with a different working directory.
    default_config_file = inteproxy.resources.default_config_file()
    parser.set_defaults(port=64609, workers=5, debug_level=0,
                        config_file=default_config_file)
    parser.add_option("--logfile")
    parser.add_option("--allow-shutdown", action="store_true")
    parser.add_option("--port", type="int")
    parser.add_option("--workers", type="int")
    parser.add_option("--debug-level", type="int")
    parser.add_option("--config-file")
    opts, rest = parser.parse_args()

    HandlerClass.allow_shutdown = opts.allow_shutdown
    HandlerClass.debug_level = opts.debug_level

    inteproxy.transcoder.transcoder_map.read_config(opts.config_file)

    setup_language()

    setup_logging(opts)

    setup_urllib2(opts.debug_level)

    server_address = ('localhost', opts.port)

    sys.stderr.write("InteProxy Version %s\n" % inteproxy_version)
    sys.stderr.write("[%s] server starting up\n" % log_date_time_string())

    httpd = ServerClass(server_address, HandlerClass, opts.workers)

    # import the gtkapp here instead of at top-level to avoid loading
    # the gtk module.  The gtk module requires an Xserver connection
    # which is undesirable when running testcases.
    from inteproxy.gtkapp import InteProxyApplication
    global the_application
    the_application = InteProxyApplication(httpd)

    print "Serving HTTP on port", opts.port, "..."
    sys.stderr.write("[%s] serving HTTP on port %d\n"
                     % (log_date_time_string(), opts.port))
    the_application.run()
    logmessage = "[%s] server stopped" % (log_date_time_string(),)
    print logmessage
    sys.stderr.write(logmessage + "\n")
